"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var bsert_1 = tslib_1.__importDefault(require("bsert"));
var elliptic_1 = tslib_1.__importDefault(require("elliptic"));
var bn_js_1 = tslib_1.__importDefault(require("bn.js"));
var hash_js_1 = tslib_1.__importDefault(require("hash.js"));
var hmac_drbg_1 = tslib_1.__importDefault(require("hmac-drbg"));
var signature_1 = require("./signature");
var curve = elliptic_1.default.ec('secp256k1').curve;
// Public key is a point (x, y) on the curve.
// Each coordinate requires 32 bytes.
// In its compressed form it suffices to store the x co-ordinate
// and the sign for y.
// Hence a total of 33 bytes.
var PUBKEY_COMPRESSED_SIZE_BYTES = 33;
/**
 * Hash (r | M).
 * @param {Buffer} msg
 * @param {BN} r
 *
 * @returns {Buffer}
 */
exports.hash = function (q, pubkey, msg) {
    var sha256 = hash_js_1.default.sha256();
    var totalLength = PUBKEY_COMPRESSED_SIZE_BYTES * 2 + msg.byteLength; // 33 q + 33 pubkey + variable msgLen
    var Q = q.toArrayLike(Buffer, 'be', 33);
    var B = Buffer.allocUnsafe(totalLength);
    Q.copy(B, 0);
    pubkey.copy(B, 33);
    msg.copy(B, 66);
    return new bn_js_1.default(sha256.update(B).digest('hex'), 16);
};
/**
 * sign
 *
 * @param {Buffer} msg
 * @param {Buffer} key
 * @param {Buffer} pubkey
 * @param {Buffer} pubNonce?
 *
 * @returns {Signature}
 */
exports.sign = function (msg, key, pubkey) {
    var prv = new bn_js_1.default(key);
    var drbg = exports.getDRBG(msg, key);
    var len = curve.n.byteLength();
    var sig;
    while (!sig) {
        var k = new bn_js_1.default(drbg.generate(len));
        sig = exports.trySign(msg, prv, k, pubkey);
    }
    return sig;
};
/**
 * trySign
 *
 * @param {Buffer} msg
 * @param {BN} prv - private key
 * @param {BN} k - DRBG-generated random number
 * @param {Buffer} pn - optional
 * @param {Buffer)} pubKey - public key
 *
 * @returns {Signature | null =>}
 */
exports.trySign = function (msg, prv, k, pubKey) {
    if (prv.isZero()) {
        throw new Error('Bad private key.');
    }
    if (prv.gte(curve.n)) {
        throw new Error('Bad private key.');
    }
    // 1a. check that k is not 0
    if (k.isZero()) {
        return null;
    }
    // 1b. check that k is < the order of the group
    if (k.gte(curve.n)) {
        return null;
    }
    // 2. Compute commitment Q = kG, where g is the base point
    var Q = curve.g.mul(k);
    // convert the commitment to octets first
    var compressedQ = new bn_js_1.default(Q.encodeCompressed());
    // 3. Compute the challenge r = H(Q || pubKey || msg)
    var r = exports.hash(compressedQ, pubKey, msg);
    var h = r.clone();
    if (h.isZero()) {
        return null;
    }
    if (h.eq(curve.n)) {
        return null;
    }
    // 4. Compute s = k - r * prv
    // 4a. Compute r * prv
    var s = h.imul(prv);
    // 4b. Compute s = k - r * prv mod n
    s = k.isub(s);
    s = s.umod(curve.n);
    if (s.isZero()) {
        return null;
    }
    return new signature_1.Signature({ r: r, s: s });
};
/**
 * Verify signature.
 *
 * @param {Buffer} msg
 * @param {Buffer} signature
 * @param {Buffer} key
 *
 * @returns {boolean}
 *
 * 1. Check if r,s is in [1, ..., order-1]
 * 2. Compute Q = sG + r*kpub
 * 3. If Q = O (the neutral point), return 0;
 * 4. r' = H(Q, kpub, m)
 * 5. return r' == r
 */
exports.verify = function (msg, signature, key) {
    var sig = new signature_1.Signature(signature);
    if (sig.s.gte(curve.n)) {
        throw new Error('Invalid S value.');
    }
    if (sig.r.gt(curve.n)) {
        throw new Error('Invalid R value.');
    }
    var kpub = curve.decodePoint(key);
    var l = kpub.mul(sig.r);
    var r = curve.g.mul(sig.s);
    var Q = l.add(r);
    var compressedQ = new bn_js_1.default(Q.encodeCompressed());
    var r1 = exports.hash(compressedQ, key, msg);
    if (r1.gte(curve.n)) {
        throw new Error('Invalid hash.');
    }
    if (r1.isZero()) {
        throw new Error('Invalid hash.');
    }
    return r1.eq(sig.r);
};
exports.toSignature = function (serialised) {
    var r = serialised.slice(0, 64);
    var s = serialised.slice(64);
    return new signature_1.Signature({ r: r, s: s });
};
/**
 * Schnorr personalization string.
 * @const {Buffer}
 */
exports.alg = Buffer.from('Schnorr+SHA256  ', 'ascii');
/**
 * Instantiate an HMAC-DRBG.
 *
 * @param {Buffer} msg
 * @param {Buffer} priv - used as entropy input
 * @param {Buffer} data - used as nonce
 *
 * @returns {DRBG}
 */
exports.getDRBG = function (msg, priv, data) {
    var pers = Buffer.allocUnsafe(48);
    pers.fill(0);
    if (data) {
        bsert_1.default(data.length === 32);
        data.copy(pers, 0);
    }
    exports.alg.copy(pers, 32);
    return new hmac_drbg_1.default({
        hash: hash_js_1.default.sha256,
        entropy: priv,
        nonce: msg,
        pers: pers,
    });
};
/**
 * Generate pub+priv nonce pair.
 *
 * @param {Buffer} msg
 * @param {Buffer} priv
 * @param {Buffer} data
 *
 * @returns {Buffer}
 */
exports.generateNoncePair = function (msg, priv, data) {
    var drbg = exports.getDRBG(msg, priv, data);
    var len = curve.n.byteLength();
    var k = new bn_js_1.default(drbg.generate(len));
    while (k.isZero() && k.gte(curve.n)) {
        k = new bn_js_1.default(drbg.generate(len));
    }
    return Buffer.from(curve.g.mul(k).encode('array', true));
};
//# sourceMappingURL=schnorr.js.map